#include "system.h"
#include "display.h"
#include "h1nchiz.h"
#include <iostream>

#define max_prime_twists 97
#define max_prime_nqs 197

bigint val(bigint p, bigint a) {
  bigint val=0;  bigint aa=a;
  if (a==0)
    return(-1);
  while(aa % p == 0) {
    val++;
    aa /= p;
  }
//  Vector<bigint> factors,exps;  Factor(factors,exps,a);
//  for(int i=1; i<=factors.size(); i++) {
//    if(factors[i] == p)
//      val=exps[i];
//  }
  return(val);
}

int omega(bigint u) {
  return(to_int(((u*u-1)/8 % 2))) ;
}

int parity(bigint a) {
  if(a%2 ==0)
    return(1);
  else
    return(-1);
}

int parity(int a) {
  if(a%2 ==0)
    return(1);
  else
    return(-1);
}

int wildhilbertone(bigint a, bigint b) {
  bigint alpha = val(2,a) ;
  if( alpha % 2 == 1) 
    return(parity(omega(b)));
  else
    return(1); 
} 

int epsilon(bigint u) {
  return(to_int(((u-1)/2) % 2 )) ;
}

int wildhilberttwo(bigint a, bigint b) {
  return( parity(epsilon(a)*epsilon(b))) ;
} 
  
int wildhilbert(bigint a, bigint b) {
  bigint alpha = val(2,a);
  bigint beta  = val(2,b);
  bigint x; power(x, 2, alpha);
  bigint aprime = a / x;
  bigint y; power(y,2,beta);
  bigint bprime = b / y;

  return(   wildhilbertone( x, bprime) 
          * wildhilbertone( y, aprime) 
          * wildhilberttwo( aprime, bprime) ) ;
}

int hilbert(bigint p, bigint a, bigint b) {
  bigint alpha = val(p,a) ;
  bigint beta  = val(p,b) ;
  bigint x; power(x,p,alpha);
  bigint aprime = a / x;
  bigint y; power(y,p,beta);
  bigint bprime = b / y ; 

  if(p==2)
    return(wildhilbert(a,b));
  else {
    bigint x; power(x,jacobi(bprime,p),alpha % 2);
    bigint y; power(y,jacobi(aprime,p),beta % 2);
    return(to_int(parity(alpha * beta * (p-1) / 2) * x * y));
  }
}

bigint Ddivisor(int i, Vector<bigint> &qstars) {
// consider i as a bitvector defining a subset of qstars
// and return the product of this subset
  bigint Di=1;
  for(int j=0; j<=qstars.size()-1; j++) {
    if(! (i % (mypower(2,j+1)) > (mypower(2,j)-1)))
      continue;
    Di*=qstars[j+1];
  }
  if(Di % 64 == 0)  Di/=16;
  return(Di);
}

void charpoly_frob(Vector<bigrational> &pol, const int p,
                   const Matrix<bigrational> &Tp, const Character &chi,
		   int k)
{
  Matrix<bigrational> Fr,A,B,C,D;
  int g=Tp.ncols();  int level=chi.modulus();
  if(level % p != 0) {       // unramified prime
    A.init(g,g); C.idmat(g); D=Tp;
    B.idmat(g);  B*= -(bigpower(p,k-1)*chi.of(p).to_bigrational());
    Fr.blocks(A,B,C,D);  Fr.charpoly(pol);
  } else {                       // ramified prime
    pol.init(0);
    A=Tp; B.init(g,g); C.init(g,g);
    if(! Tp.iszero()) {
      Tp.Inverse(D);  D *= (bigpower(p,k-1));
      Fr.blocks(A,B,C,D);  Fr.charpoly(pol);
    }
  }
}

int next_point(int i, Vector<bigrational> &pol) {
  if(i<=0)
    return(-1);
  i--;
  while(pol.polcoeff(i) == 0 && i>=0)
    i--;
  return(i);
}

void newton_poly(const Vector<bigrational> &pol, int p) {
  Vector<bigrational> npol = pol;

//  cout << "Newton poly = ";
//  for(int i=npol.poldeg(); i>=0; i--) 
//    cout << val(p,npol.polcoeff(i).numerator()) << " ";
//  cout << endl;

  bool done = false;

  while (! done) {
    bigrational slope = -999999;
    bigrational next_slope = slope;
    done = true;
    for(int i=npol.poldeg(); i>=0; i=next_point(i,npol))  {
      int j=next_point(i,npol);
      if(j<0)  continue;
      next_slope = val(p,npol.polcoeff(j).numerator()) -
                     val(p,npol.polcoeff(i).numerator());
      next_slope /= (i-j);
      if(next_slope <= slope) {
        npol[i+1] = 0;
        done = false;
        break;
      }
      slope = next_slope;
    }
  }

  cout << "Slopes: ";
  for(int i=npol.poldeg(); i>=0; i=next_point(i,npol)) {
    int j=next_point(i,npol);
    if(j<0) continue;
    bigrational slope = val(p,npol.polcoeff(j).numerator()) -
                         val(p,npol.polcoeff(i).numerator());
    slope /= (i-j);
    cout << slope << "[" << i-j << "] ";
  }
  cout << endl;
}

void chooseform(Matrix<bigrational> &V, H1NChiZ *H)
{
  cout << "dim V = " << V.ncols() << endl;

  MatVecList<bigrational> E;
  H->BasisOfEigenforms(E,V,300,2);
  cout << E.len() << " eigenforms found." << endl;

  for(int i=1; i<=E.len(); i++) {
    cout << i << ": dim = ";
    cout << E.M(i-1).ncols() << endl;
  }
  cout << endl;

  int m;
  do {
    cout << "Select an eigenform: ";
    cin >> m; 
    if(!cin) exit(-1);
  } while(m<0 || m>E.len());

  Matrix<bigrational> K=E.M(m-1);

/*
  int n;
  cout << "Factor of Tn,  n = ";
  cin >> n;
  if(!cin)   exit(-1);
  if(n<=0)   exit(-1);

  cout << "Choose one of the following factors."<<endl;
  Matrix<bigrational> T;
  H->Tn_On(n,V,T);
  Vector<bigrational> pol;
  T.charpoly(pol);
  VecFac<bigrational> fac;
  pol.polfactor(fac);
  for(int i=1;i<=fac.numfactors();i++) {
    cout << "\t" << i << ": " ; 
    if(fac.exp(i)>1) cout << "(" ;
    fac.base(i).polprint();
    if(fac.exp(i)>1) cout << ")^" << fac.exp(i);
    cout << endl;
  }
  cout << "\t" << fac.numfactors()+1 << ": ALL factors"<<endl;
  int m;
  do {
    cout << "Select a factor: ";
    cin >> m;
    if(m<=0) {n=0; exit(-1);}
  } while (m>fac.numfactors()+1);
  Vector<bigrational> f;
  if(m==fac.numfactors()+1) 
    f=pol;
  else {
//          fac.base(m).polpower(f, fac.exp(m));
    f = fac.base(m);
  }
  cout << "Selected factor = "; f.polprint();
  cout << endl;

  Matrix<bigrational> K;
  H->Tn_Eig_On(n,V,f,K); 
*/

//   ostrstream ost(Vname,1024);
//    cout << "ker(f(T"<<n<<")|previous), f=";
//    fr.polprint(cout);
//    cout << endl;
//    cout << "Setting V=ker(f(T" << n << ")), dim="
//         << V.ncols() << "." << endl;

  V=K;
}


main(int argc, char *argv[]) {

//  Messages.off();

  int k;
  cout << "Weight = ";
  cin >> k;
  if(!cin)  exit(-1);

  bool save=true;
  Character chi(save);
  int N=chi.modulus();
  bool plus=true;
  H1NChiZ* H = new H1NChiZ(chi,k,save,plus);
  Matrix<bigrational> V = H->get_Sknew();

// select an eigenform
  chooseform(V,H);

// find all the twists

  bool check_twists = get_bool("Check twisting?");

  Vector<bigint> qstars,exps;  Factor(qstars,exps,N);
  for(int i=1; i<=qstars.size(); i++) {
    if(qstars[i] % 2 != 0)
      qstars[i] *= parity((qstars[i]-1)/2);
  }
  if(qstars[1] == 2) {   // I am assuming increasing order of prime factors !!
    qstars.resize(qstars.size()+1);
    qstars.shift(1);
    qstars[1] = 8;
    qstars[2] = -8;
  }

  int numtwists=0;
  Vector<int> twists;  twists.init(mypower(2,qstars.size()));
  for(int i=0; i<=mypower(2,qstars.size())-1; i++) {
    bool twist=true;
    if(! check_twists) {
      twists[i+1]=1;  numtwists++;
      continue;
    }
    bigint Di = Ddivisor(i,qstars);
//    cout << "i = " << i << endl << "Di = " << Di << endl;
    for(int p=3; p<=max_prime_twists; p=next_prime(p)) {
      if(jacobi(Di,p) == -1) {
        Matrix<bigrational> T;
        H->Tn_On(p,V,T);
        Vector<bigrational> pol;
        T.charpoly(pol);
        for(int i=2; i<=pol.size(); i+=2) {
          if(pol[i]!=0) {
            twist=false;
            break;
          }
        }
      }
      if(twist == false)
        break;
    }
    if(twist == true) { 
//      cout << "Factor " << Di << " is a twist baby" << endl;
      twists[i+1]=1;
      numtwists++;
    } else {
      twists[i+1]=0;
    }
  }
  cout << "Twists: ";
  for(int i=0; i<=mypower(2,qstars.size())-1; i++)  {
    if(twists[i+1] == 0)  continue;
    bigint Di=Ddivisor(i,qstars);
    cout << Di << " ";
  }
  cout << endl << "numtwists = " << numtwists << endl;

  if(numtwists > V.ncols())  {
    cout << "Something is wrong (CM form?) : numtwists > [E:Q]!!" << endl;
    exit(-1);
  }

// compute hilbert symbols

// Step 1 -- find the nq's and compute zq's.

  cout << endl << "Computing nq's and zq's" << endl;
  Vector<int> nqs;  nqs.init(qstars.size());
  Vector<bigint> zqs;  zqs.init(qstars.size());
  int q=0;
  bigrational ProdNorms=1;
  for (int j=1; j<=qstars.size(); j++) {
    int qstar=to_int(qstars[j]);
    int foundnq=0; int passed=0;
    for (int nq=3; nq<=max_prime_nqs; nq=next_prime(nq)) {
      for(int i=0; i<=mypower(2,qstars.size())-1; i++) {
        if(twists[i+1]==0)  continue;
        bigint Di = Ddivisor(i,qstars);
        passed=0;
        if(qstar % 2 != 0) {
          if(Di % qstar == 0  && jacobi(Di,nq) != -1)
            break;
          if(Di % qstar != 0 && jacobi(Di,nq) != 1)
            break;
          passed=1;
        } else {
          int j= qstar>0 ? 0 : 1;
          if(i % (mypower(2,j+1)) > (mypower(2,j)-1) && jacobi(Di,nq) != -1)
            break;
          if(!(i % (mypower(2,j+1)) > (mypower(2,j)-1)) && jacobi(Di,nq) != 1)
            break;
          passed=1;
        }
      }
      if(passed) {
        if(N % nq == 0)  continue;
        Matrix<bigrational> T,T2;
        H->Tn_On(nq,V,T);
        if(T.iszero())  continue;
//        cout << "T" << nq << ": "; DisplayCharPoly(T); cout << endl;
        nqs[j] = nq;  foundnq=1; 
        cout << "n" << qstar << " = " << nq << endl;

// compute Norm from F of zq
        T2=T*T;
//        cout << "T^2" << ": "; DisplayCharPoly(T2); cout << endl;
        Vector<bigrational> pol; T2.charpoly(pol);
        VecFac<bigrational> fac; pol.polfactor(fac);
//        fac.base(1).polprint(); cout << endl;
        bigrational Norm;
        Norm=fac.base(1).polcoeff(0); Norm *= parity(fac.base(1).poldeg());
	// This is the absolute norm of zq.
        int I= T.ncols(); I/=numtwists; I /= fac.base(1).poldeg();
        Norm = bigpower(Norm.numerator(),I);
        // This is the norm from F of zq
        cout << "Norm from F of z" << qstar << " = " << Norm << endl;
        ProdNorms *= Norm;
        zqs[j] = Norm.numerator();
        break;
      }
    }
    if (!foundnq ) {
      cout << "Could not find n" << qstar << "!  Exiting." << endl;
      exit(-1);
    }
  }

// Step 2 -- compute local Hilbert symbols at all primes testp
//             dividing 2*N*norms(zq's)

  cout << endl << "Computing local Hilbert symbols" << endl;
  Vector<bigint> BadFactors,BadExps;
  Factor(BadFactors,BadExps,2*N*ProdNorms.numerator());
  for(int i=1; i<=BadFactors.size(); i++) {
    bigint testp=BadFactors[i];
    int ProdHilbs=1;
    for(int i=1; i<=qstars.size(); i++) {
      bigint qstar = qstars[i];  
//      if(qstar%2 != 0)  qstar *= parity((qstar-1)/2);
      int Hilb = hilbert(testp,zqs[i],qstar);
      ProdHilbs *= Hilb;
      cout << "Norm zq = " << zqs[i] << "; q* = " << qstar <<
           "; p = " << testp << "; Hilbert = " << Hilb << endl;
    }
    cout << "Product = " << ProdHilbs << endl;

    if(testp < 999999) {
      Matrix<bigrational> T;
      H->Tn_On(to_int(testp),V,T);
      Vector<bigrational> pol;
      charpoly_frob(pol,to_int(testp),T,chi,k);
//      cout << "charpoly(Fr) = "; pol.polprint(); cout << endl;
      newton_poly(pol,to_int(testp));
      if (T.iszero())
        cout << "T_p = 0 !!!" << endl;
      cout << endl;
    } else {
      cout << "p is too big to compute Newton polygon" << endl;
    }
  }
  delete H;
}
