#include <iostream>
#include "system.h"
#include "h1nchi.h"
#include "NTL/ZZ.h"

int a = 1;
int m = 11;

template <class F>
void checknonordinary(int p, Vector<F> ap, Vector<F> heckefield,int orbit) {
  return;
}

template <>
void checknonordinary(int p, Vector<bigrational> ap, 
			Vector<bigrational> heckefield, int orbit) {

// bigrational norm = ap.polcoeff(0) * ap.polcoeff(0) - ap.polcoeff(0) * 
//           ap.polcoeff(1) - ap.polcoeff(1) * ap.polcoeff(1);
//  cout << "  ap: "; ap.polprint(cout);
//  cout << "  x: "; heckefield.polprint(cout);

  bigrational norm = pol_resultant(heckefield,ap);
  norm /= p;
  if (norm.denominator() == 1) {
    norm *= p;
    cout << endl << "NON-ORDINARY prime" << endl ;
    cout << "p: " << p;
    cout << "  orbit: " << orbit; 
    cout << "  a_p: ";    ap.polprint(cout);
    cout << "  x: ";   heckefield.polprint(cout);
    cout << "  norm(a_p): " << norm << endl;
  }
  return;
}

template <class F>
bool
HeckeEigenvalues_ap(H1NChi<bigrational> *H, const Matrix<F>& V, int startp, int maxp) 
{
    // 1. Find a dual basis of eigenforms using DualBasisOfEigenforms
    
    // 2. Systematically find a very simple modular symbol e such
    //    that <e,f> =/= 0 for all f in the dual basis.  Save the
    //    nonzero values <e,f> once they are found. 
    
    // 3. Without first computing T_p on the whole space, compute
    //    e_p:=T_p(e) as quickly as possible.  This is the MAIN
    //    bottleneck in the computation. 
    
    // 4. Record the quotients a_p=<e_p,f>/<e,f>
    
    // Step 1.
    MatVecList<F> W;
    if(!H->DualBasisOfEigenforms(W, V)) return false;
    cout << "dim V = " << V.ncols()<<endl;
    cout << "num eigenforms found = " << W.len()<<endl;

    for(int i=0; i<W.len(); i++) {
      cout << "orbit: " << i+1 << "  dim: " << W.v(i).poldeg() << endl;
    }
    
    // Step 2.
    mesg("Finding e.\n");
    
    // Find a vector of length dim_m, as simple as possible, so that it is
    // *not* in the kernel of matrix multiplication on the left
    // by any of the W[i].  The entries of e are either 1 or 0.
    int dim_m = H->dim_Mk();
    Vector<int> e; e.init(dim_m); 
    // Step 2a. quickly try vectors with exactly one nonzero entry
    bool e_simple=false;		// e_simple means we e is simple.
    int e_simple_pos;
    Vector<F> ef[W.len()];		
    for(int i=1;i<=dim_m;i++) {
	bool all_nonzero=true;
	for(int j=0;j<W.len();j++) {
	    W.M(j).row(i,ef[j]);
	    if(ef[j].iszero()) { all_nonzero=false; break; }
	}
	if (all_nonzero) {
	    e[i]=1; e_simple=true; e_simple_pos=i;
	    break; 
	}
    }
    
    // Step 2b. 2a didn't find anything, so use a systematic (but slower) search.
    if(!e_simple) {
	mesg("No monomial e. Trying randomized strategy: ");
	// (i) Find an e not orthogonal to any W[i]
	//     Using a slow search here makes sense because choosing
	//     a good e is crucial to the speed of the rest of the algorithm. 
	int i=2;
	bool not_done=true;
	while(i<=dim_m && not_done) {
	    // try a few hundred random vectors of length i.
	    for(int j=0;j<300;j++) {
		int a[i];
		for(int n=0;n<i;n++)
		    a[n]=RandomBnd(dim_m)+1;
		bool all_nonzero=true;
		for(int n=0;n<W.len();n++) {
		    Vector<F> row;
		    ef[n].init(W.M(n).ncols());
		    for(int m=0;m<i;m++) {
			W.M(n).row(a[m],row);
			ef[n] += row;
		    }
		    if(ef[n].iszero()) { all_nonzero=false; break; }
		}
		if(all_nonzero) {			   // found it!
		    for(int n=0;n<i;n++)
			e[a[n]]=1;
		    not_done=false;
		    break;
		}
	    }
	    if(not_done) i++;
	}
	
	mesg("Randomized strategy suceeded.");
	// (ii) find all pairings ef
	// ... this is automatic in the above algorithm. 
	
    }
    
    // Compute and save the values 1/<e,f>.
    mesg("Computing the values 1/<e,f>. [");
    Vector<F> oneover_ef[W.len()];
    for(int i=0;i<W.len();i++) {
	reset_timer();
	oneover_ef[i] = ef[i].polinv_polymod(W.v(i));
	mesg((i*100)/W.len(),"% ");
	mesg("(");mesg(get_time());mesg("s)");
    }
    mesg("]\n");

    if(! is_prime(startp)) startp=next_prime(startp);
    int numprimes=0;
    for (int p=startp; p<=maxp; p=next_prime(p)) 
	if(is_prime(p)) numprimes++;
    
//    aplist.init(W.len());
//    for(int i=0;i<W.len();i++) {
//	aplist.M(i).init(numprimes,W.v(i).poldeg());
//	aplist.v(i)=W.v(i);
//    }
    int pidx=1;
    mesg("Computing ap:  [");
    for (int p=startp; p<=maxp; p=next_prime(p)) {
        // mesg(p," ");
	// Step 3. 
	// Compute T_p(e). 
	Vector<F> Tpe;
	Array<int> Heilbronn;
	H->Tp_Heilbronn (p, Heilbronn) ;
	
	if(e_simple) {
	    // Compute T_p ( e_simple_pos -th generating Manin symbol )
	    H->Tp_StandardSymbol(Tpe,Heilbronn,e_simple_pos);
	} else {
	    Tpe.init(dim_m);
	    for(int i=1;i<=e.size();i++) 
		if(e[i]) {
		    Vector<F> Tpei;
		    H->Tp_StandardSymbol(Tpei,Heilbronn,i);
		    Tpe+=Tpei;
		}
	}
	
	// Step 4. 
	// Compute ap
	for(int i=0; i<W.len(); i++) {
	    Vector<F> Tpef = Tpe * W.M(i);
	    Vector<F> ap=polmul_polymod(Tpef,oneover_ef[i],W.v(i));
checknonordinary(p,ap,W.v(i),i+1);

if(p % m == a){ 
cout << "p = " << p 
     << ", a_p mod p = " << ap[1].numerator() % p 
     << ", jacobi = " << jacobi(ap[1].numerator(), p) << endl; 
} 
//	    aplist.M(i).setrow(pidx,ap);
	}
	pidx++;
    }

    mesgln("]");
    return true;
}

main() {

//  Messages.off();

  int level=1;
  int startp=2;  int maxp=100;
  int k=12;
  bool save=true;
  Array<int> degs(1);  degs[0]=1;

//  Character chi(save);

//  cout << "Level = ";
//  cin >> level;
//  if(!cin)  exit(-1) ;
//  cout << "startp = ";
//  cin >> startp;
//  if(!cin)  exit(-1) ;
  cout << "modulus = ";
  cin >> m;
  cout << "congruence class = ";
  cin >> a;

  cout << "maxp = ";
  cin >> maxp;
  if(!cin)  exit(-1) ;

  Character chi(level,1,save,&degs);

  H1NChi<bigrational>* H=new H1NChi<bigrational>(chi,k,save);
  const Matrix<bigrational>& V = H->get_Sk();
  if(V.ncols()==0)  { 
    cout << "Empty space" << endl;
    exit(-1);
  }
  HeckeEigenvalues_ap(H,V,startp,maxp);
  delete H;
}
